//
//  Winmail.m
//  TNEFsEnough3
//
//  Created by Josh Jacob on 5/27/11.
//  Copyright 2011 Josh Jacob. All rights reserved.
//

#import "Winmail.h"
#import "FileTableItem.h"
#import "FileTableViewDataSource.h"
#import "ytnef.h"
#import "PrintTNEF.h"

@implementation Winmail

@synthesize from;
@synthesize subject;
@synthesize date;
@synthesize body;
@synthesize bodyHTML;
@synthesize fileDataSource;
@synthesize fieldFrom;
@synthesize fieldSubject;
@synthesize fieldDate;
@synthesize tableFiles;
@synthesize fieldBody;
@synthesize fieldBodyScrollView;
@synthesize webview;
@synthesize webViewBox;
@synthesize bodyIsHTML;
@synthesize exportFile;
@synthesize exportDir;
@synthesize exportArray;
@synthesize outputStream;
@synthesize savingSheet;
@synthesize savingSheetLabel;
@synthesize savingSheetProgress;

- (id)init
{
    self = [super init];
    if (self) {
        // Add your subclass-specific initialization here.
        // If an error occurs here, send a [self release] message and return nil.
        
        cStringPath = nil;
        
        exportArray = [[NSMutableArray alloc] initWithCapacity: 1];
        
        bodyIsHTML = NO;
    }
    return self;
}

- (void)awakeFromNib
{
    [self.tableFiles setDraggingSourceOperationMask:NSDragOperationCopy forLocal:NO];
    [self.tableFiles registerForDraggedTypes:[NSArray arrayWithObject: NSFilesPromisePboardType] ];
}

- (void) dealloc
{
    // ytnef
    TNEFFree(&tnefStruct);
    
    // free
    if (cStringPath != nil)
        free(cStringPath);
    
    // release datasource
    self.fileDataSource = nil;
    self.exportArray = nil;
    self.outputStream = nil;
    
    // release
    self.from = nil;
    self.subject = nil;
    self.date = nil;
    self.body = nil;
    
    [super dealloc];
}

- (NSString *)windowNibName
{
    // Override returning the nib file name of the document
    // If you need to use a subclass of NSWindowController or if your document supports multiple NSWindowControllers, you should remove this method and override -makeWindowControllers instead.
    return @"Winmail";
}

#pragma mark - NSDocument

- (void)windowControllerDidLoadNib:(NSWindowController *)aController
{
    [super windowControllerDidLoadNib:aController];
    // Add any code here that needs to be executed once the windowController has loaded the document's window.
    
    [self.tableFiles setDataSource: self.fileDataSource];
    
    if (bodyIsHTML) 
    {
        [[self.webview mainFrame] loadHTMLString:self.bodyHTML baseURL:nil];
        [self.fieldBodyScrollView setHidden:YES];
        [self.webViewBox setHidden:NO];
        [self.webview setHidden:NO];
    }
}

- (NSData *)dataOfType:(NSString *)typeName error:(NSError **)outError {
    /*
     Insert code here to write your document to data of the specified type. If outError != NULL, ensure that you create and set an appropriate error when returning nil.
    You can also choose to override -fileWrapperOfType:error:, -writeToURL:ofType:error:, or -writeToURL:ofType:forSaveOperation:originalContentsURL:error: instead.
    */
    if (outError) {
        *outError = [NSError errorWithDomain:NSOSStatusErrorDomain code:unimpErr userInfo:NULL];
    }
    return nil;
}

- (BOOL)readFromURL:(NSURL *)absoluteURL ofType:(NSString *)typeName error:(NSError **)outError
{
    // get file path as c string
    NSString * pathString = [absoluteURL path];
    cStringPath = malloc(sizeof(char) * [pathString length] + 2);
    [pathString getCString:cStringPath maxLength:([pathString length] + 2) encoding:NSUTF8StringEncoding];
    
    // init tnef structure
    TNEFInitialize(&tnefStruct);
    #ifdef DEBUG
    tnefStruct.Debug = 1;
    #endif
    
    // parse file
    int parseResult = TNEFParseFile(cStringPath, &tnefStruct);
    if (parseResult < 0) {
        DLog(@"ERROR processing file\n");
        NSDictionary * errorUserInfo = nil;
        /*
             #define YTNEF_CANNOT_INIT_DATA -1
             #define YTNEF_NOT_TNEF_STREAM  -2
             #define YTNEF_ERROR_READING_DATA -3
             #define YTNEF_NO_KEY -4
             #define YTNEF_BAD_CHECKSUM -5
             #define YTNEF_ERROR_IN_HANDLER -6
             #define YTNEF_UNKNOWN_PROPERTY -7
             #define YTNEF_INCORRECT_SETUP -8
         */
        switch (parseResult) {
            case YTNEF_CANNOT_INIT_DATA:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Cannot initialize data.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_NOT_TNEF_STREAM:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"This is not a TNEF file.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_ERROR_READING_DATA:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Error reading file.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_NO_KEY:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"No TNEF key found.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_BAD_CHECKSUM:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Bad checksum.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_ERROR_IN_HANDLER:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Error in handler.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_UNKNOWN_PROPERTY:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Unknown property.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            case YTNEF_INCORRECT_SETUP:
                errorUserInfo = [NSDictionary dictionaryWithObjectsAndKeys: @"Incorrect setup.", NSLocalizedFailureReasonErrorKey, nil];
                break;
            default:
                break;
        }
        NSError * err = [NSError errorWithDomain:@"TNEF's Enough Error Domain." code:parseResult userInfo:errorUserInfo];
        *outError = err;
        return NO;
    }
    
    // set from/subject/date fields
    if (tnefStruct.from.size > 0) {
        from = [[NSString alloc] initWithCString: tnefStruct.from.data encoding:NSWindowsCP1252StringEncoding];
    } else {
        variableLength * hasSender = MAPIFindProperty(&tnefStruct.MapiProperties, PROP_TAG(PT_UNICODE, PR_SENDER_EMAIL_ADDRESS));
        if (hasSender != MAPI_UNDEFINED) {
            from = [[NSString alloc] initWithCString: hasSender->data encoding:NSWindowsCP1252StringEncoding];
        }
    }
    if (tnefStruct.subject.size > 0) {
        subject = [[NSString alloc] initWithCString: tnefStruct.subject.data encoding:NSWindowsCP1252StringEncoding];
    }
    if (tnefStruct.dateSent.wYear > 0) {
        NSDateComponents * comps = [[NSDateComponents alloc] init];
        [comps setHour:tnefStruct.dateSent.wHour];
        [comps setMinute:tnefStruct.dateSent.wMinute];
        [comps setSecond:tnefStruct.dateSent.wSecond];
        [comps setDay:tnefStruct.dateSent.wDay];
        [comps setMonth:tnefStruct.dateSent.wMonth];
        [comps setYear:tnefStruct.dateSent.wYear];
        NSCalendar *gregorian = [[NSCalendar alloc] initWithCalendarIdentifier:NSGregorianCalendar];
        self.date = [gregorian dateFromComponents:comps];
        [gregorian release];
        [comps release];
    }
    
    // set body
    [fieldBody insertText:@"Test"];
    variableLength * hasBody = MAPIFindProperty(&tnefStruct.MapiProperties, PROP_TAG(PT_BINARY, PR_RTF_COMPRESSED));
    if (hasBody != MAPI_UNDEFINED) {
        if (IsCompressedRTF(hasBody) == 1)
        {
            int size;
            unsigned char * rawBody = DecompressRTF(hasBody, &size);

            NSData *data;
            
            data = [[NSData alloc] initWithBytes:rawBody length:size];
            body = [[NSAttributedString alloc] initWithRTF:data documentAttributes:NULL];
            
            [data release];
            free(rawBody);
            
        } else {
            
        }
    } else {
        
        variableLength * hasBody = MAPIFindProperty(&tnefStruct.MapiProperties, PROP_TAG(PT_BINARY, PR_BODY_HTML));
        if (hasBody != MAPI_UNDEFINED) {

            bodyIsHTML = YES;
            NSData *data;
            data = [[NSData alloc] initWithBytes:hasBody->data length:hasBody->size];
            
            NSString * htmlString = [[NSString alloc] initWithData:data encoding:NSISOLatin1StringEncoding];
            self.bodyHTML = htmlString;
            
            [htmlString release];
            [data release];
        }
    }
    
    // get embedded files
    
    fileDataSource = [[FileTableViewDataSource alloc] init];
    
    Attachment *p = tnefStruct.starting_attach.next;
    while (p != NULL) {

        int i; //, object;
        char ifilename[256];
        variableLength *filename;
//        variableLength *filedata;
        
        if  (p->RenderData.atyp == 1)
        {
            // case 1: printf("File      "); break;
            
//            if (p->Title.size > 0) 
//                DLog("%s", p->Title.data);
//            DLog("\n");
            
//            if (p->RenderData.dwFlags == 0x00000001) 
//                printf("     MAC Binary Encoding\n");
//            if (p->TransportFilename.size >0) 
//                printf("     Transported under the name %s\n", p->TransportFilename.data);
            
//            if (p->Date.wYear >0 ) {
//                printf("    Date: ");
//                TNEFPrintDate(p->Date);
//                printf("\n");
//            }
//            if (p->CreateDate.wYear > 0) {
//                printf("    Creation Date: ");
//                TNEFPrintDate(p->CreateDate);
//                printf("\n");
//            }
//            if (p->ModifyDate.wYear > 0) {
//                printf("    Modified on: ");
//                TNEFPrintDate(p->ModifyDate);
//                printf("\n");
//            }
            
//            if (p->MAPI.count>0) {
//                printf("    MAPI Properties: %i\n", p->MAPI.count);
//                //if (verbose == 1) {
//                //    MAPIPrint(&p->MAPI);
//                //}
//            }
            
            if (p->FileData.size > 0) {
                // DLog("    Attachment Size:  %ib\n", p->FileData.size);
                
                if ((filename = MAPIFindProperty(&(p->MAPI), PROP_TAG(30,0x3707))) == MAPI_UNDEFINED) {
                    if ((filename = MAPIFindProperty(&(p->MAPI), PROP_TAG(30,0x3001))) == MAPI_UNDEFINED) {
                        filename = &(p->Title);
                    }
                }
                
                sprintf(ifilename, "%s", filename->data);
                for(i=0; i<strlen(ifilename); i++) 
                    if (ifilename[i] == ' ') 
                        ifilename[i] = '_';
                // DLog("    File saves as [%s]\n", ifilename);
                
                
                // fill in table item
                FileTableItem * fileItem = [[FileTableItem alloc] init];
                fileItem.numBytes = p->FileData.size;
                fileItem.attachment = p;
                if (p->Title.size > 0) {
                    NSString * filenameNS = [[NSString alloc] initWithCString:ifilename encoding:NSWindowsCP1252StringEncoding];
                    fileItem.fileName = filenameNS;
                    filenameNS = nil;
                }
                if (fileItem.fileName == nil) {
                    fileItem.fileName = @"Untitled Attachment";
                }
                [self.fileDataSource.fileArray addObject: fileItem];
                [fileItem release];
                
            }
        }
        p=p->next;
    }
    
    #ifdef DEBUG
    PrintTNEF(tnefStruct);
    #endif
    
    return YES;
}

- (BOOL)readFromData:(NSData *)data ofType:(NSString *)typeName error:(NSError **)outError {
    /*
    Insert code here to read your document from the given data of the specified type. If outError != NULL, ensure that you create and set an appropriate error when returning NO.
    You can also choose to override -readFromFileWrapper:ofType:error: or -readFromURL:ofType:error: instead.
    */
    if (outError) {
        *outError = [NSError errorWithDomain:NSOSStatusErrorDomain code:unimpErr userInfo:NULL];
    }
    return YES;
}

- (BOOL)validateUserInterfaceItem:(id <NSValidatedUserInterfaceItem>)anItem
{
    SEL theAction = [anItem action];
    
    if (theAction == @selector(exportSelectedAttachment:))
    {
        if ([self.tableFiles numberOfSelectedRows] > 0)
        {
            return YES;
        }
        return NO;
    }
    
    if (theAction == @selector(exportAll:))
    {
        if ([self.tableFiles numberOfRows] > 0)
        {
            return YES;
        }
        return NO;
    }
    
    return [super validateUserInterfaceItem:anItem];
}

#pragma mark - IBActions

- (IBAction) exportSelectedAttachment: (id) sender
{
    // test if more than one selected
    if ([self.tableFiles numberOfSelectedRows] > 1) {
        [self exportSet: [self.fileDataSource.fileArray objectsAtIndexes:[self.tableFiles selectedRowIndexes]]];
        return;
    }
    
    // get filename
    FileTableItem * filetableItem = [self.fileDataSource.fileArray objectAtIndex:[self.tableFiles selectedRow]];
    NSString * filename = filetableItem.fileName;
    
    NSWindow * window = [[[self windowControllers] objectAtIndex:0] window];
        
    // Set the default name for the file and show the panel.
    NSSavePanel * panel = [NSSavePanel savePanel];
    [panel setNameFieldStringValue: filename];
    [panel beginSheetModalForWindow:window completionHandler:^(NSInteger result) {
        if (result == NSFileHandlingPanelOKButton)
        {
            // setup single file export
            exportOne = YES;
            self.exportFile = [panel URL];
            [self.exportArray removeAllObjects];
            [self.exportArray addObject:filetableItem];
            DLog(@"Will export to file: %@", self.exportFile);
            
            [self performSelectorOnMainThread:@selector(kickOffExport) withObject:nil waitUntilDone:NO];
        }
    }];
    
}

- (void) kickOffExport
{
    // show saving sheet
    [self.savingSheetProgress startAnimation: self];
    [NSApp beginSheet:self.savingSheet 
       modalForWindow:[self windowForSheet] 
        modalDelegate:self 
       didEndSelector:nil 
          contextInfo:nil];
    
    // display filename
    FileTableItem * filetableItem = [self.exportArray lastObject];
    NSString * filename = filetableItem.fileName;
    NSString * newLabel = [NSString stringWithFormat:@"Saving file \"%@\"...", filename];
    [self.savingSheetLabel setStringValue:newLabel];
    [self.exportArray removeLastObject];
    
    // create output stream
    self.outputStream = [[NSOutputStream alloc] initWithURL:self.exportFile append:NO];
    [self.outputStream setDelegate:self];
    [self.outputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
    [self.outputStream open];
}

- (IBAction) exportAll: (id) sender
{
    [self exportSet: self.fileDataSource.fileArray];
}

- (void) exportSet: (NSArray *) rows
{
    // get the main window for the document.
    NSWindow* window = [[[self windowControllers] objectAtIndex:0] window];
    
    // create and configure the panel.
    NSOpenPanel* panel = [NSOpenPanel openPanel];
    [panel setCanChooseDirectories:YES];
    [panel setCanChooseFiles:NO];
    [panel setAllowsMultipleSelection:NO];
    [panel setCanCreateDirectories:YES];
    [panel setMessage:@"Select export directory."];
    
    // display the panel attached to the document's window.
    [panel beginSheetModalForWindow:window completionHandler:^(NSInteger result){
        if (result == NSFileHandlingPanelOKButton) {
            NSArray* urls = [panel URLs];
            
            // setup export to dir
            exportOne = NO;
            self.exportDir = [urls objectAtIndex:0];;
            DLog(@"Will export to dir: %@", self.exportDir);
            
            // add all file to array
            for (FileTableItem * fileTableItem in rows) {
                [self.exportArray addObject:fileTableItem];
            }
            
            [self performSelectorOnMainThread:@selector(kickOffExportAll) withObject:nil waitUntilDone:NO];
        }
        
    }];
}

- (void) kickOffExportAll
{
    // show saving sheet
    [self.savingSheetProgress startAnimation: self];
    [NSApp beginSheet:self.savingSheet 
       modalForWindow:[self windowForSheet] 
        modalDelegate:self 
       didEndSelector:nil 
          contextInfo:nil];
    
    [self kickOffExportAllNext];
}

- (void) kickOffExportAllNext
{
    // NSFileManager * fileManager = [[NSFileManager alloc] init];

    // get filename
    FileTableItem * filetableItem = [self.exportArray lastObject];
    NSString * filename = filetableItem.fileName;
    NSString * newLabel = [NSString stringWithFormat:@"Saving file \"%@\"...", filename];
    [self.savingSheetLabel setStringValue:newLabel];
    
    // get unique name
    NSString * destFile  = [FileTableViewDataSource GenerateUniqueFileNameAtPath: [self.exportDir path] WithBase:[filename stringByDeletingPathExtension] Extension:[filename pathExtension]];
    NSURL * destFileUrl = [NSURL fileURLWithPath:destFile];
    DLog(@"Saving file %@", destFile);
    
    // create output stream
    self.outputStream = [[NSOutputStream alloc] initWithURL:destFileUrl append:NO];
    [self.outputStream setDelegate:self];
    [self.outputStream scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
    [self.outputStream open];
    
    // [fileManager release];
}

#pragma mark - NSStreamDelegate

- (void)stream:(NSStream *)theStream handleEvent:(NSStreamEvent)streamEvent
{
    if (theStream == self.outputStream)
    {
        NSOutputStream * theOutputStream = (NSOutputStream *) theStream;
        
        switch(streamEvent)
        {
            case NSStreamEventHasSpaceAvailable:
            {
                // get reference to export file and attachment
                FileTableItem * filetableItem;
                if (exportOne == YES) {
                    filetableItem = [self.fileDataSource.fileArray objectAtIndex:[self.tableFiles selectedRow]];
                } else {
                    filetableItem = (FileTableItem *) [self.exportArray lastObject];
                }
                Attachment * attachment = filetableItem.attachment;
                
                // write out data
                NSNumber * size = [NSNumber numberWithInt:attachment->FileData.size];
                uint8_t * data = (uint8_t *) attachment->FileData.data;
                NSInteger len = [theOutputStream write: data maxLength: [size unsignedLongValue]];
                DLog(@"Written length: %ld", len);
                
                // close up current stream
                [theOutputStream close];
                [theOutputStream removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
                [theOutputStream release];
                self.outputStream = nil;
                
                // remove last object; look for more
                if (exportOne == NO) {
                    [self.exportArray removeLastObject];
                }
                if ([self.exportArray count] > 0) {
                    [self kickOffExportAllNext];
                } else {
                    [NSApp performSelectorOnMainThread:@selector(endSheet:) withObject:self.savingSheet waitUntilDone:NO];
                    [self.savingSheet orderOut: self];
                }
             
                break;
            }
        }
        
    }
}

@end
